# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional

import jax
import jax.numpy as jnp
import torch
from jax.sharding import Mesh, NamedSharding, PartitionSpec
from torch.nn.parameter import Parameter
from torchax.interop import jax_view, torch_view
from torchax.ops.mappings import t2j
from vllm.attention.layer import Attention
from vllm.model_executor.layers.fused_moe.layer import (
    FusedMoE, FusedMoEConfig, UnquantizedFusedMoEMethod)
from vllm.model_executor.layers.linear import (LinearBase,
                                               UnquantizedLinearMethod)
from vllm.model_executor.layers.quantization import \
    register_quantization_config
from vllm.model_executor.layers.quantization.base_config import (
    QuantizationConfig, QuantizeMethodBase)

from tpu_inference.layers.common.quant_methods import (UNQUANTIZED,
                                                       get_tpu_quant_method)
from tpu_inference.layers.common.sharding import ShardingAxisName
from tpu_inference.layers.common.utils import \
    slice_sharded_tensor_for_concatenation
from tpu_inference.layers.vllm.fused_moe import (FusedMoEBackend,
                                                 fused_moe_apply,
                                                 select_moe_backend)
from tpu_inference.layers.vllm.process_weights.fused_moe_weights import (
    FusedMoEWeights, process_moe_weights, shard_moe_weights)
from tpu_inference.layers.vllm.process_weights.linear_weights import (
    LinearWeights, process_lienar_weights, shard_linear_weights,
    to_parameter_list)
from tpu_inference.layers.vllm.quantization.configs import (
    VllmQuantConfig, VllmQuantLinearConfig)
from tpu_inference.logger import init_logger
from tpu_inference.utils import get_mesh_shape_product

P = PartitionSpec

logger = init_logger(__name__)


@register_quantization_config(get_tpu_quant_method(UNQUANTIZED))
class VllmUnquantizedConfig(QuantizationConfig, VllmQuantConfig):

    @classmethod
    def get_name(cls) -> str:
        return UNQUANTIZED

    @classmethod
    def get_supported_act_dtypes(cls) -> list[torch.dtype]:
        return [torch.float32, torch.float16, torch.bfloat16]

    @classmethod
    def get_min_capability(cls) -> int:
        return 0  # Always supported

    @classmethod
    def get_config_filenames(cls) -> list[str]:
        return []  # No extra configs required.

    @classmethod
    def from_config(cls, _: dict[str, Any]) -> "VllmUnquantizedConfig":
        return cls()

    def get_quant_method(self, layer: torch.nn.Module,
                         prefix: str) -> Optional[QuantizeMethodBase]:
        if isinstance(layer, LinearBase):
            linear_config = self.get_linear_config(layer)
            return VllmUnquantizedLinearMethod(linear_config)
        if isinstance(layer, FusedMoE):
            moe_config = self.get_moe_config(layer)
            return VllmUnquantizedFusedMoEMethod(moe_config, self.mesh)
        if isinstance(layer, Attention):
            return None
        return None


class VllmUnquantizedLinearMethod(UnquantizedLinearMethod):

    def __init__(self, linear_config: VllmQuantLinearConfig):
        self.linear_config = linear_config

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        weight = t2j(layer.weight, use_dlpack=False)
        delattr(layer, "weight")
        if layer.bias is not None and not layer.skip_bias_add:
            if layer.return_bias:
                logger.warning_once("Bias might return incorrect value.")
            bias = t2j(layer.bias, use_dlpack=False)
            delattr(layer, "bias")
        else:
            bias = None

        @jax.jit
        def process_unquantized_linear_weights(
            weight: jax.Array,
            bias: jax.Array | None,
        ) -> LinearWeights:
            return process_lienar_weights(
                LinearWeights(
                    weight=weight,
                    weight_scale=None,
                    zero_point=None,
                    bias=bias,
                ),
                fused=self.linear_config.fuse_matmuls,
                output_sizes=self.linear_config.output_sizes,
                reorder_size=self.linear_config.n_shards,
            )

        weights = process_unquantized_linear_weights(weight, bias)
        weights = torch_view(
            shard_linear_weights(
                weights,
                mesh=self.linear_config.mesh,
                weight_p_spec=self.linear_config.weight_sharding,
                bias_p_spec=self.linear_config.bias_sharding,
            ))

        if self.linear_config.fuse_matmuls:
            layer.weight = Parameter(weights.weight, requires_grad=False)
            if bias is not None:
                layer.bias = Parameter(weights.bias, requires_grad=False)
        else:
            layer.weight = to_parameter_list(weights.weight)
            if bias is not None:
                layer.bias = to_parameter_list(weights.bias)

    def apply(self,
              layer: torch.nn.Module,
              x: torch.Tensor,
              bias: Optional[torch.Tensor] = None) -> torch.Tensor:
        assert isinstance(layer, LinearBase)

        with jax.named_scope(layer._get_name()):
            if in_sharding := self.linear_config.get_input_sharding(x):
                x.shard_(NamedSharding(self.linear_config.mesh, in_sharding))

            if self.linear_config.fuse_matmuls:
                out = self._apply_fused(layer, x, bias)
            else:
                out = self._apply_split(layer, x, bias)

            if out_sharding := self.linear_config.get_output_sharding(out):
                out.shard_(NamedSharding(self.linear_config.mesh,
                                         out_sharding))

        return out

    def _apply_fused(self,
                     layer: torch.nn.Module,
                     x: torch.Tensor,
                     bias: Optional[torch.Tensor] = None) -> torch.Tensor:
        x_jax = jax_view(x)
        weight_jax = jax_view(layer.weight)

        outs = jnp.einsum("mn,pn->mp", x_jax, weight_jax)
        if bias is not None and not layer.skip_bias_add:
            outs += bias.jax()

        outs = slice_sharded_tensor_for_concatenation(
            outs, self.linear_config.output_sizes, self.linear_config.n_shards)
        out = jnp.concatenate(outs, axis=-1)
        return torch_view(out)

    def _apply_split(self,
                     layer: torch.nn.Module,
                     x: torch.Tensor,
                     bias: Optional[torch.Tensor] = None) -> torch.Tensor:
        assert isinstance(layer.weight, torch.nn.ParameterList)

        x_jax = x.jax()
        outs = []
        for i, weight in enumerate(layer.weight):
            weight_jax = jax_view(weight)

            out = jnp.einsum("mn,pn->mp", x_jax, weight_jax)
            if bias is not None and not layer.skip_bias_add:
                out += jax_view(bias[i])

            outs.append(out)
        out = jnp.concatenate(outs, axis=-1)
        return torch_view(out)


class VllmUnquantizedFusedMoEMethod(UnquantizedFusedMoEMethod):

    def __init__(
        self,
        moe: FusedMoEConfig,
        mesh: Mesh,
        ep_axis_name: str = "model",
    ):
        super().__init__(moe)
        self.mesh = mesh
        self.moe_backend = select_moe_backend(self.moe)

        self.extra_backend_kwargs = {}
        if self.moe_backend == FusedMoEBackend.FUSED_MOE:
            # When fused moe kernle is used, we pass extra arguments like
            # tuned block sizes to the kernel.
            self.extra_backend_kwargs = dict(
                ep_axis_name=ep_axis_name,
                # TODO: Use autotune table once we have it.
                bt=64,
                bf=1024,
                bd1=1536,
                bd2=1536,
                btc=64,
                bfc=1024,
                bd1c=1536,
                bd2c=1536,
            )

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:
        assert isinstance(layer, FusedMoE)

        w13_weight = t2j(layer.w13_weight, use_dlpack=False)
        w2_weight = t2j(layer.w2_weight, use_dlpack=False)

        if self.moe.has_bias:
            w13_bias = t2j(layer.w13_bias, use_dlpack=False)
            w2_bias = t2j(layer.w2_bias, use_dlpack=False)
        else:
            w13_bias = w2_bias = None

        @jax.jit
        def process_unquantized_moe_weights(
            w13_weight: jax.Array,
            w13_bias: jax.Array | None,
            w2_weight: jax.Array,
            w2_bias: jax.Array | None,
        ) -> FusedMoEWeights:

            w13_interleave = layer.activation == "swigluoai"
            w13_reorder_size = get_mesh_shape_product(
                self.mesh, ShardingAxisName.MLP_TENSOR)

            return process_moe_weights(
                FusedMoEWeights(
                    w13_weight=w13_weight,
                    w13_weight_scale=None,
                    w13_bias=w13_bias,
                    w2_weight=w2_weight,
                    w2_weight_scale=None,
                    w2_bias=w2_bias,
                ),
                moe_backend=self.moe_backend,
                w13_reorder_size=w13_reorder_size,
                w13_interleave=w13_interleave,
            )

        weights = process_unquantized_moe_weights(
            w13_weight,
            w13_bias,
            w2_weight,
            w2_bias,
        )
        weights = torch_view(
            shard_moe_weights(weights, self.moe_backend, self.mesh))

        layer.w13_weight = Parameter(weights.w13_weight, requires_grad=False)
        layer.w2_weight = Parameter(weights.w2_weight, requires_grad=False)

        if self.moe.has_bias:
            layer.w13_bias = Parameter(weights.w13_bias, requires_grad=False)
            layer.w2_bias = Parameter(weights.w2_bias, requires_grad=False)

    def apply(
        self,
        layer: torch.nn.Module,
        x: torch.Tensor,
        router_logits: torch.Tensor,
    ) -> torch.Tensor:

        return fused_moe_apply(
            layer,
            x,
            router_logits,
            self.moe_backend,
            self.mesh,
            self.extra_backend_kwargs,
        )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from typing import Optional

import jax
import pytest
import torch
import torchax
from jax.sharding import PartitionSpec
from torchax.interop import torch_view
from torchax.ops.mappings import j2t, t2j
from vllm.config import set_current_vllm_config
from vllm.distributed.parallel_state import (ensure_model_parallel_initialized,
                                             init_distributed_environment)
from vllm.engine.arg_utils import EngineArgs
from vllm.model_executor.layers.linear import (ColumnParallelLinear,
                                               LinearBase,
                                               MergedColumnParallelLinear,
                                               QKVParallelLinear,
                                               RowParallelLinear)
from vllm.model_executor.layers.quantization.utils.quant_utils import \
    pack_quantized_values_into_int32
from vllm.model_executor.model_loader import get_model as vllm_get_model
from vllm.scalar_type import scalar_types

from tpu_inference.layers.vllm.quantization import get_tpu_quantization_config
from tpu_inference.layers.vllm.quantization.awq import (VllmAWQConfig,
                                                        VllmAWQLinearMethod)
from tpu_inference.layers.vllm.quantization.configs import \
    VllmQuantLinearConfig

from . import utils as test_utils

P = PartitionSpec
MODELS = ["Qwen/Qwen2.5-1.5B-Instruct-AWQ"]


def ref_quantize_uint4(x: torch.Tensor, group_size: int):
    uint4_max = 15

    # For group quantization, we reshape so that x[0], x[1], ... x[i] are
    # quantized with different scale values.
    x = torch.reshape(x, (-1, group_size) + (x.shape[1:]))

    # Equation for asymmetric quantization is x_q = (x + x_z) / scale where
    # x_z is calculated to ensure x + x_z does not contain any negative values.
    offset = torch.clamp(-torch.amin(x, dim=1, keepdim=True), min=0)
    x += offset
    # After adding offset, x will not contain any negative values.
    assert x.min() >= 0

    x_abs_max = torch.amax(x, dim=1, keepdim=True)
    x_s = x_abs_max / uint4_max
    # torch does not support uint4, therefore, we cast to int32 instead.
    x_q = torch.clip(x / x_s, 0, uint4_max).to(torch.int32)
    x_z = torch.clip(offset / x_s, 0, uint4_max).to(torch.int32)
    return x_q, x_z, x_s.to(torch.float32)


def ref_w4a16(x: torch.Tensor, w_q: torch.Tensor, w_z: torch.Tensor,
              w_s: torch.Tensor, b: Optional[torch.Tensor]):
    # Dequantize asymetric quantized weight.
    w = (w_q.to(torch.float32) - w_z.to(torch.float32)) * w_s
    w = w.reshape((-1, w.shape[-1]))
    out = torch.einsum('bd,df->bf', x.to(torch.float32), w)
    if b is not None:
        out += b
    return out.to(x.dtype)


def pack_awq_weight_into_int32(weight: torch.Tensor):
    # AWQ packs 8 uint4 into 32-bits in this order.
    awq_order = (0, 2, 4, 6, 1, 3, 5, 7)

    orig_shape = weight.shape
    weight = weight.reshape(orig_shape[:-1] + (-1, 8))
    weight = weight[..., awq_order].reshape(orig_shape)

    return pack_quantized_values_into_int32(weight, scalar_types.uint4, 1)


def return_ref_and_layer_output(
    layer: torch.nn.Module,
    qweight: torch.Tensor,
    qzeros: torch.Tensor,
    scales: torch.Tensor,
    batch_size: int = 16,
):
    assert isinstance(layer, LinearBase)
    quant_method = layer.quant_method
    assert isinstance(quant_method, VllmAWQLinearMethod)
    quant_config = quant_method.quant_config
    assert isinstance(quant_config, VllmAWQConfig)
    jax_config = quant_method.linear_config
    assert isinstance(jax_config, VllmQuantLinearConfig)

    input_tensor = torch.rand(
        batch_size, layer.input_size, dtype=torch.bfloat16) / 10
    input_tensor = input_tensor.to('cpu')

    ref_output = ref_w4a16(
        input_tensor,
        qweight,
        qzeros,
        scales,
        layer.bias,
    )

    # Run torchax/jax function
    quant_method.process_weights_after_loading(layer)
    with torchax.default_env():
        jax_input_tensor = torch_view(t2j(input_tensor, use_dlpack=False))
        layer_output = layer(jax_input_tensor)
        layer_output = j2t(layer_output.to(torch.float32)).to(torch.bfloat16)

    return ref_output, layer_output


def initialize_and_return_layer_weights(layer: torch.nn.Module):
    assert isinstance(layer, LinearBase)
    quant_method = layer.quant_method
    assert isinstance(quant_method, VllmAWQLinearMethod)
    quant_config = quant_method.quant_config
    assert isinstance(quant_config, VllmAWQConfig)
    jax_config = quant_method.linear_config
    assert isinstance(jax_config, VllmQuantLinearConfig)

    # torch.rand returns value in the range of [0, 1). We subtract by 0.2 to
    # simulate asymmetry
    weight = torch.rand((layer.input_size, layer.output_size)) - 0.2
    qweight, qzeros, scales = ref_quantize_uint4(weight,
                                                 quant_config.group_size)

    # We modify uint4 quantized weights into AWQ format.
    layer_qweight = qweight.reshape((-1, layer.output_size))
    layer_qzeros = qzeros.reshape((-1, layer.output_size))
    layer_scales = scales.reshape((-1, layer.output_size))

    layer_qweight = pack_awq_weight_into_int32(layer_qweight)
    layer_qzeros = pack_awq_weight_into_int32(layer_qzeros)

    assert layer.qweight.data.shape == layer_qweight.shape
    assert layer.qzeros.data.shape == layer_qzeros.shape
    assert layer.scales.data.shape == layer_scales.shape

    layer.qweight.data = layer_qweight
    layer.qzeros.data = layer_qzeros
    layer.scales.data = layer_scales

    bias = None
    if layer.bias is not None:
        bias = torch.rand_like(layer.bias.data)
        layer.bias.data = bias

    return qweight, qzeros, scales, bias


@pytest.fixture(autouse=True)
def setup_environment():
    # This is a fake config used for init dist env.
    # RowParallelLinear needs dist env to be initialized.
    engine_args = EngineArgs(
        model=MODELS[0],
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )

    vllm_config = engine_args.create_engine_config()

    with set_current_vllm_config(vllm_config):
        temp_file = tempfile.mkstemp()[1]
        init_distributed_environment(
            1,
            0,
            local_rank=0,
            distributed_init_method=f"file://{temp_file}",
            backend="gloo")
        ensure_model_parallel_initialized(1, 1)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
def test_quant_override(model, mesh):

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.model_config.dtype = torch.bfloat16

    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    assert isinstance(quant_config, VllmAWQConfig)
    assert quant_config.vllm_config == vllm_config
    assert quant_config.mesh == mesh


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize(
    "mesh",
    [
        test_utils.get_spmd_mesh(1),
        # We limit device count by 2 instead of using all devices (like 8) since
        # AWQ requires n_groups to be divisible by number of shards. Qwen uses
        # group size of 128 and one of the layer has input size of 1536, meaning
        # n_groups = 1536//128 = 12 - which is not divisible by 8.
        test_utils.get_spmd_mesh(min(jax.local_device_count(), 2))
    ])
def test_loading_model(model, mesh):
    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.model_config.dtype = torch.bfloat16
    vllm_config.quant_config = get_tpu_quantization_config(vllm_config, mesh)
    vllm_config.device_config.device = "cpu"

    vllm_model = vllm_get_model(vllm_config=vllm_config)
    layers = test_utils.find_all_layer_type(vllm_model, LinearBase)
    for layer in layers:
        assert isinstance(layer.quant_config, VllmAWQConfig)
        assert isinstance(layer.quant_method, VllmAWQLinearMethod)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
@pytest.mark.parametrize("enable_sp", [False, True])
def test_row_parallel_linear(model, bias, mesh, enable_sp):
    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    vllm_config.model_config.dtype = dtype
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = RowParallelLinear(
            input_size=4096,
            output_size=8192,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    qweight, qzeros, scales, _ = initialize_and_return_layer_weights(
        linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(
        linear_layer, qweight, qzeros, scales)
    torch.testing.assert_close(ref_output, layer_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
@pytest.mark.parametrize("enable_sp", [False, True])
def test_column_parallel_linear(model, bias, mesh, enable_sp):
    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    # Call tpu_inference code
    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = ColumnParallelLinear(
            input_size=4096,
            output_size=8192,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    qweight, qzeros, scales, _ = initialize_and_return_layer_weights(
        linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(
        linear_layer, qweight, qzeros, scales)
    torch.testing.assert_close(ref_output, layer_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("fuse_matmuls", [False, True])
def test_qkv_parallel_linear(model, bias, mesh, enable_sp, fuse_matmuls):
    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    # Call tpu_inference code
    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = QKVParallelLinear(
            hidden_size=4096,
            head_size=128,
            total_num_heads=32,
            total_num_kv_heads=8,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )
        linear_layer.quant_method.fuse_matmuls = fuse_matmuls

    qweight, qzeros, scales, _ = initialize_and_return_layer_weights(
        linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(
        linear_layer, qweight, qzeros, scales)
    torch.testing.assert_close(ref_output, layer_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
@pytest.mark.parametrize("fuse_matmuls", [False, True])
@pytest.mark.parametrize("enable_sp", [False, True])
def test_merged_column_parallel_linear(model, bias, mesh, fuse_matmuls,
                                       enable_sp):
    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    # Call tpu_inference code
    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        linear_layer = MergedColumnParallelLinear(
            input_size=4096,
            output_sizes=[14336] * 2,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )
        linear_layer.quant_method.fuse_matmuls = fuse_matmuls

    qweight, qzeros, scales, _ = initialize_and_return_layer_weights(
        linear_layer)
    ref_output, layer_output = return_ref_and_layer_output(
        linear_layer, qweight, qzeros, scales)
    torch.testing.assert_close(ref_output, layer_output)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import jax
import jax.numpy as jnp
import numpy as np
import pytest
import torch
from flax import nnx
from jax.sharding import Mesh
from vllm.config import ModelConfig

# Assuming the model file is named deepseek_v3.py
from tpu_inference.models.jax.deepseek_v3 import (DeepSeekV3,
                                                  DeepSeekV3WeightLoader)


class MockVariable:
    """Mocks an nnx.Variable or a QArray structure."""

    def __init__(self, shape, dtype=jnp.bfloat16, sharding=None):
        self.value = jnp.zeros(shape, dtype=dtype)
        self.sharding = sharding or (None, ) * len(shape)
        self.nbytes = self.value.nbytes
        # Handle the QArray structure used in the loader
        self.array = SimpleNamespace(
            qvalue=self,
            scale=SimpleNamespace(
                value=jnp.ones((1, )),
                nbytes=4,
                sharding=None,
                addressable_shards=[SimpleNamespace(data=jnp.ones((1, )))]))
        self.addressable_shards = [SimpleNamespace(data=self.value)]


class MockVllmConfig:
    """Mock VllmConfig for DeepSeekV3."""

    def __init__(self,
                 model_name: str = "deepseek-ai/DeepSeek-V3",
                 use_mla: bool = False):
        self.model_config = MagicMock(spec=ModelConfig)
        self.model_config.model = model_name
        self.model_config.use_mla = use_mla

        # DeepSeek V3 specific config
        hf_config = MagicMock()
        hf_config.num_hidden_layers = 1  # Small for testing
        hf_config.num_nextn_predict_layers = 1
        self.model_config.hf_config = hf_config

        self.load_config = MagicMock()
        self.load_config.download_dir = None

        self.cache_config = MagicMock()
        self.cache_config.cache_dtype = "auto"

        self.additional_config = {
            "random_weights": False,
            "sparse_matmul": False,
            "is_verbose": True
        }


@pytest.fixture(scope="module")
def mesh():
    if not jax.devices():
        pytest.skip("No JAX devices available.")
    devices = np.array(jax.local_devices())
    num_devices = len(devices)
    device_mesh = devices.reshape((num_devices, 1, 1, 1))
    # Simplify axis names for testing
    with Mesh(device_mesh,
              axis_names=('data', 'attn_dp', 'model', 'expert')) as m:
        yield m


@pytest.fixture
def rng():
    return jax.random.PRNGKey(0)


@pytest.fixture
def mock_config():
    return MockVllmConfig()


class TestDeepSeekV3:

    def test_init(self, mock_config, rng, mesh):
        """Tests if the model initializes with the correct hierarchy."""
        model = DeepSeekV3(mock_config, rng, mesh)
        assert len(model.layers) == 3  # num_layers from mock
        assert isinstance(model.embedder, nnx.Module)
        assert model.vllm_config.model_config.hf_config.num_hidden_layers == 1

    def test_random_weights(self, mock_config, rng, mesh):
        """Tests that force_random_weights initializes non-zero weights."""
        with jax.set_mesh(mesh):
            model = DeepSeekV3(mock_config,
                               rng,
                               mesh,
                               force_random_weights=True)
            # Check embedding
            weight = model.embedder.input_embedding_table_VD.value
            assert jnp.std(weight) > 0
            # Check a layer norm (should be 1s usually, but check existence)
            assert model.final_norm.scale.value.shape == (7168, )

    @patch("tpu_inference.models.jax.deepseek_v3.DeepSeekV3WeightLoader")
    def test_load_weights_called(self, mock_loader_cls, mock_config, rng,
                                 mesh):
        model = DeepSeekV3(mock_config, rng, mesh)
        mock_loader_instance = mock_loader_cls.return_value

        model.load_weights(rng)

        mock_loader_instance.load_weights.assert_called_once_with(model)


class TestDeepSeekV3WeightLoader:

    @pytest.fixture
    def loader(self, mock_config):
        # We need to mock the generator so it doesn't try to download files
        with patch(
                "tpu_inference.models.jax.deepseek_v3.model_weights_generator",
                return_value=[]):
            return DeepSeekV3WeightLoader(vllm_config=mock_config,
                                          num_layers=2,
                                          hidden_size=7168,
                                          q_lora_rank=1536,
                                          kv_lora_rank=512,
                                          attn_heads=128,
                                          qk_nope_head_dim=128,
                                          qk_rope_head_dim=64,
                                          v_head_dim=128,
                                          num_local_experts=256,
                                          model_dtype=jnp.bfloat16)

    @pytest.mark.parametrize("loaded_key, expected_mapped", [
        ("model.embed_tokens.weight", "embedder.input_embedding_table_VD"),
        ("model.layers.0.self_attn.q_a_proj.weight",
         "layers.0.attn.kernel_q_down_proj_DA"),
        ("model.layers.5.mlp.experts.10.gate_proj.weight",
         "layers.5.custom_module.kernel_gating_EDF"),
        ("model.layers.1.mlp.shared_experts.down_proj.weight",
         "layers.1.shared_experts.kernel_down_proj_FD"),
        ("model.norm.weight", "final_norm.scale"),
    ])
    def test_key_mapping(self, loader, loaded_key, expected_mapped):
        assert loader.map_loaded_to_standardized_name(
            loaded_key) == expected_mapped

    def test_transpose_params(self, loader):
        # Test a standard MLP transpose (1, 0)
        dummy_weight = jnp.ones((100, 200))
        transposed = loader._transpose_params("mlp.down_proj", dummy_weight)
        assert transposed.shape == (200, 100)

        # Test MLA kernel transpose (2, 0, 1)
        dummy_mla = jnp.ones((10, 20, 30))
        transposed_mla = loader._transpose_params("k_b_proj", dummy_mla)
        assert transposed_mla.shape == (30, 10, 20)

    def test_moe_stacking_logic(self, loader):
        """Tests that individual expert weights are collected and stacked correctly."""
        weights_dict = {}
        layer_num = "0"
        loader.num_routed_experts = 4  # Small for test

        # Simulate loading 4 experts
        for i in range(4):
            name = f"model.layers.0.mlp.experts.{i}.gate_proj.weight"
            weight = torch.ones((10, 20)) * i
            result = loader._process_moe_weights(name, weight, weights_dict)

            if i < 3:
                assert result is None
                assert weights_dict[layer_num][1] == i + 1
            else:
                # On the last expert, it should return stacked tensor
                assert result is not None
                assert result.shape == (4, 10, 20)
                assert layer_num not in weights_dict  # Should be cleaned up

    def test_mla_kernel_weight_splitting(self, loader, mesh):
        """Tests that kv_b_proj is split into k_b_proj and v_b_proj for MLA kernel."""
        loader.use_mla_kernel = True
        loader.attn_heads = 2
        loader.qk_nope_head_dim = 4
        loader.v_head_dim = 4
        loader.kv_lora_rank = 8

        # Total rows = heads * (nope_dim + v_dim) = 2 * (4 + 4) = 16
        # Cols = kv_lora_rank = 8
        kv_b_proj_weight = torch.randn((16, 8))

        # Mocking the load_individual_weight to capture what gets passed
        with patch.object(loader,
                          '_load_individual_weight',
                          return_value=(0, 0)):
            model_mock = MagicMock()
            model_mock.mesh = mesh

            # Simulate the splitting logic in the loader
            weight_reshaped = kv_b_proj_weight.view(2, 4 + 4, 8)
            k_weight = weight_reshaped[:, :4, :]
            v_weight = weight_reshaped[:, 4:, :]

            # Verify shapes of split parts
            assert k_weight.shape == (2, 4, 8)
            assert v_weight.shape == (2, 4, 8)

    def test_load_individual_weight_with_mxfp4(self, loader, mesh):
        """Tests the logic for unpacking MXFP4 weights."""
        name = "layers.0.attn.kernel_q_down_proj_DA"
        # Mocking torch tensor as uint8 (packed fp4)
        expected_weight_shape = (128, 128)  # Unpacked
        expected_scale_shape = (128, 1)

        weight = torch.zeros(expected_weight_shape, dtype=torch.uint8)
        scale = torch.ones(expected_scale_shape, dtype=torch.float32)

        # Mock model parameters
        mock_var = MockVariable(
            (128, 128),
            dtype=jnp.float4_e2m1fn,
            sharding=(None, ('attn_dp', 'model',
                             'expert')))  # Unpacked shape (64 * 2)
        mock_params = {
            "layers": {
                "0": {
                    "attn": {
                        "kernel_q_down_proj_DA": mock_var
                    }
                }
            }
        }

        with patch("tpu_inference.models.jax.deepseek_v3.get_param", return_value=mock_var), \
             patch("tpu_inference.models.jax.deepseek_v3.u8_unpack_e2m1") as mock_unpack, \
             patch("jax.make_array_from_callback") as mock_make_array:

            def side_effect_router(shape, *args, **kwargs):
                if shape == expected_scale_shape:
                    # Return FP32 for the scale call
                    return jnp.ones(shape, dtype=jnp.float32)
                elif shape == expected_weight_shape:
                    # Return FP4 for the weight call
                    return jnp.zeros(shape, dtype=jnp.float4_e2m1fn)
                return jnp.zeros(shape)  # Fallback

            mock_make_array.side_effect = side_effect_router
            mock_unpack.return_value = torch.zeros(expected_weight_shape)

            loader._load_individual_weight(name,
                                           weight,
                                           mock_params,
                                           mesh,
                                           scale=scale)

            mock_unpack.assert_called_once()
            (actual_arg, ), _ = mock_unpack.call_args
            # The implementation converts the torch weight to a JAX array
            expected_arg = jnp.array(weight.cpu().numpy())
            assert jnp.array_equal(actual_arg, expected_arg).item()
            assert mock_make_array.called

    def test_load_weights_full_flow(self, loader, mesh):
        """Integrative test for the load_weights loop."""
        model = MagicMock(spec=nnx.Module)
        model.mesh = mesh

        # Setup generator to return one normal weight
        loader.names_and_weights_generator = [("model.embed_tokens.weight",
                                               torch.ones((10, 10)))]

        mock_var = MockVariable((10, 10))

        with patch("tpu_inference.models.jax.deepseek_v3.nnx.state"), \
             patch("tpu_inference.models.jax.deepseek_v3.get_param", return_value=mock_var), \
             patch("tpu_inference.models.jax.deepseek_v3.nnx.update"), \
             patch.object(loader, '_load_individual_weight', return_value=(1.0, 0.5)):

            loader.load_weights(model)
            # Verify verbose logging worked if enabled
            assert loader.is_verbose is True

    def test_load_individual_weight_unpacked(self, loader, mesh):
        """
        Tests the logic for loading 'unpacked' weights (e.g., standard FP8).
        This verifies the branch that uses DTYPE_VIEW_MAP for raw memory conversion.
        """
        name = "layers.0.attn.kernel_q_down_proj_DA"

        # 1. Setup a standard 'unpacked' FP8 torch tensor
        # DeepSeek V3 weights are often float8_e4m3fn
        weight_shape = (128, 128)
        weight = torch.randn(weight_shape).to(torch.float8_e4m3fn)

        # 2. Mock model parameters to expect jnp.float8_e4m3fn
        # We reuse the MockVariable helper but specify the dtype
        mock_var = MockVariable(weight_shape, dtype=jnp.float8_e4m3fn)
        mock_params = {
            "layers": {
                "0": {
                    "attn": {
                        "kernel_q_down_proj_DA": mock_var
                    }
                }
            }
        }

        # 3. Patch the necessary JAX/Utility functions
        with patch("tpu_inference.models.jax.deepseek_v3.get_param", return_value=mock_var), \
             patch("tpu_inference.models.jax.deepseek_v3.u8_unpack_e2m1") as mock_unpack, \
             patch("jax.make_array_from_callback") as mock_make_array:

            # Mock the JAX array creation to return a dummy
            mock_make_array.return_value = jnp.zeros(weight_shape,
                                                     dtype=jnp.float8_e4m3fn)

            # Execute the loader method
            loader._load_individual_weight(name,
                                           weight,
                                           mock_params,
                                           mesh,
                                           scale=None)

            # VERIFICATIONS:
            # - u8_unpack_e2m1 should NOT be called for standard FP8 (only for packed uint8 + scale)
            mock_unpack.assert_not_called()

            # - make_array_from_callback should be called with the correct shape and sharding
            # The first argument to make_array_from_callback is the shape
            assert mock_make_array.call_args[0][0] == weight_shape

            # - Verify the model weight value was updated (even if with our dummy)
            assert mock_var.value.dtype == jnp.float8_e4m3fn

    def test_load_individual_weight_with_scale(self, loader, mesh):
        """
        Tests loading an unpacked weight that also has a quantization scale.
        """
        name = "layers.0.custom_module.kernel_gating_DF"
        weight_shape = (64, 128)
        scale_shape = (64, 1)

        # Use BF16 for this test to verify DTYPE_VIEW_MAP handles multiple types
        weight = torch.randn(weight_shape).to(torch.bfloat16)
        scale = torch.ones(scale_shape, dtype=torch.float32)

        mock_var = MockVariable(weight_shape, dtype=jnp.bfloat16)
        mock_params = {
            "layers": {
                "0": {
                    "custom_module": {
                        "kernel_gating_DF": mock_var
                    }
                }
            }
        }

        with patch("tpu_inference.models.jax.deepseek_v3.get_param", return_value=mock_var), \
             patch("jax.make_array_from_callback") as mock_make_array:

            def side_effect_router(shape, *args, **kwargs):
                if shape == scale_shape:
                    # Return FP32 for the scale call
                    return jnp.ones(shape, dtype=jnp.float32)
                elif shape == weight_shape:
                    # Return FP4 for the weight call
                    return jnp.zeros(shape, dtype=jnp.bfloat16)
                return jnp.zeros(shape)  # Fallback

            mock_make_array.side_effect = side_effect_router

            loader._load_individual_weight(name,
                                           weight,
                                           mock_params,
                                           mesh,
                                           scale=scale)

            # Verify the scale was applied to the MockVariable's internal QArray structure
            # (In the model code: base_model_weight.array.scale.value = maybe_sharded_scale)
            assert mock_var.array.scale.value is not None



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import jax
import jax.numpy as jnp
import numpy as np
from flax import nnx
from jax.sharding import Mesh, PartitionSpec

from tpu_inference.layers.jax.moe.deepseek_v3_moe import (DeepSeekV3Router,
                                                          SparseMoE)


class TestDeepSeekV3Router(unittest.TestCase):

    def setUp(self):
        self.cpu_mesh = Mesh(jax.devices('cpu'), axis_names=('data', ))

    def test_get_topk_indices_single_group(self):
        """Test get_topk_indices with single expert group."""
        with jax.set_mesh(self.cpu_mesh):
            router = DeepSeekV3Router(random_init=True,
                                      hidden_size=512,
                                      num_experts=4,
                                      num_experts_per_tok=2,
                                      n_groups=1,
                                      topk_groups=1,
                                      norm_topk_prob=True,
                                      routed_scaling_factor=1.0,
                                      dtype=jnp.bfloat16,
                                      rngs=nnx.Rngs(42))
            router.bias_E = jnp.zeros((4, ))

            scores = jnp.array([[0.1, 0.3, 0.2, 0.4]])  # shape: (1, 4)
            indices = router.get_topk_indices(scores)

            # Should return indices of top 2 experts
            expected_indices = jnp.array([[3,
                                           1]])  # experts with scores 0.4, 0.3
            self.assertTrue(jnp.array_equal(indices, expected_indices))

    def test_get_topk_indices_2_groups(self):
        """Test get_topk_indices with 2 expert groups."""
        with jax.set_mesh(self.cpu_mesh):
            router = DeepSeekV3Router(random_init=True,
                                      hidden_size=512,
                                      num_experts=4,
                                      num_experts_per_tok=2,
                                      n_groups=2,
                                      topk_groups=1,
                                      norm_topk_prob=True,
                                      routed_scaling_factor=1.0,
                                      dtype=jnp.bfloat16,
                                      rngs=nnx.Rngs(42))
            router.bias_E = jnp.zeros((4, ))

            # 4 experts, 2 groups, 2 experts per group
            scores = jnp.array([[[0.1, 0.3, 0.2, 0.4]]])  # shape: (1, 1, 4)
            indices = router.get_topk_indices(scores)

            # Should return indices of top 2 experts
            expected_indices = jnp.array([[[3, 2]]])
            self.assertTrue(jnp.array_equal(indices, expected_indices))

    def test_router_e2e(self):
        with jax.set_mesh(self.cpu_mesh):
            router = DeepSeekV3Router(random_init=True,
                                      hidden_size=512,
                                      num_experts=8,
                                      num_experts_per_tok=2,
                                      n_groups=2,
                                      topk_groups=1,
                                      norm_topk_prob=True,
                                      routed_scaling_factor=1.0,
                                      dtype=jnp.bfloat16,
                                      rngs=nnx.Rngs(42))
            x = jnp.ones((2, 512))
            weights, indices = router(x)
            self.assertEqual(weights.shape, (2, 2))
            self.assertEqual(indices.shape, (2, 2))


class TestSparseMoE(unittest.TestCase):

    def setUp(self):
        """Set up a multi-device mesh and a sample MoE layer for testing."""
        devices = jax.devices()
        self.device_count = len(devices)
        if self.device_count < 8:
            self.skipTest("This test requires at least 8 simulated devices.")

        # This mesh will have a 'model' axis for expert parallelism
        mesh_shape = (self.device_count, 1)
        device_mesh_array = np.array(devices).reshape(mesh_shape)

        # Define the axis names
        axis_names = ('model', 'data')

        # Create the 2D mesh
        self.mesh = Mesh(device_mesh_array, axis_names=axis_names)

        # --- Model Configuration ---
        self.B, self.S, self.D = 2, 4, 16  # Batch, Sequence, Hidden Dim
        self.E, self.K = 16, 8  # Num Experts, Experts per Token
        self.moe_intermediate_size = 32  # FFN Dim
        self.num_expert_parallelism = 8  # Shard experts across 8 devices

        self.key = jax.random.PRNGKey(42)
        self.x = jax.random.normal(self.key, (self.B * self.S, self.D),
                                   dtype=jnp.bfloat16)

        # --- Instantiate MoE Layer ---
        # We need to do this inside the mesh context
        with self.mesh:
            router = DeepSeekV3Router(hidden_size=self.D,
                                      num_experts=self.E,
                                      num_experts_per_tok=self.K,
                                      n_groups=1,
                                      topk_groups=1,
                                      norm_topk_prob=False,
                                      routed_scaling_factor=1.0,
                                      dtype=jnp.bfloat16,
                                      rngs=nnx.Rngs(self.key),
                                      ed_sharding=PartitionSpec(),
                                      e_sharding=PartitionSpec(),
                                      activation_ffw_td=PartitionSpec(
                                          'data', None))
            # Instantiation updated to match user's code snippet
            self.moe = SparseMoE(
                hidden_size=self.D,
                intermediate_size_moe=self.moe_intermediate_size,
                num_local_experts=self.E,
                hidden_act="silu",
                num_experts_per_tok=self.K,
                router=router,
                dtype=jnp.bfloat16,
                rngs=nnx.Rngs(self.key),
                mesh=self.mesh,
                apply_expert_weight_before_computation=False,

                # Sharding specs updated based on user's snippet
                edf_sharding=PartitionSpec('model', None, None),
                efd_sharding=PartitionSpec('model', None, None),
                activation_ffw_ted=PartitionSpec('data', None),
                activation_ffw_td=PartitionSpec(
                    'data', None)  # Activations are replicated
            )

    def test_token_replicated_expert_parallel_fwd(self):
        """
        Validates the MoE forward pass against a simple, dense equivalent.
        This specifically tests the is_batch_sharded_by_expert=False path.
        """
        # --- 1. Get the ACTUAL output from the complex distributed MoE layer ---
        # The __call__ method will trigger the shard_map, which requires the mesh context.
        with self.mesh:
            actual_output = self.moe(self.x)

        # --- 2. Calculate the EXPECTED output using a simple, sequential process ---
        # This serves as the "ground truth".

        # Get router decisions (router params are replicated, so this is fine)
        router_weights, selected_experts = self.moe.router(self.x)

        # Gather the full, unsharded weights from all devices ---
        # .value on a sharded param gives the *local* shard.
        # jax.device_get() retrieves the *full* GlobalDeviceArray to the host.
        gating_kernel_full = jax.device_get(self.moe.kernel_gating_EDF.value)
        up_proj_kernel_full = jax.device_get(self.moe.kernel_up_proj_EDF.value)
        down_proj_kernel_full = jax.device_get(
            self.moe.kernel_down_proj_EFD.value)

        # Check that we really got the full weights
        self.assertEqual(gating_kernel_full.shape,
                         (self.E, self.D, self.moe_intermediate_size))

        # Flatten inputs for easier iteration
        flat_x = self.x.reshape(self.B * self.S, self.D)
        flat_weights = router_weights.reshape(self.B * self.S, self.K)
        flat_experts = selected_experts.reshape(self.B * self.S, self.K)

        expected_output = jnp.zeros_like(flat_x)

        # Manually apply each expert to each token sequentially
        for i in range(self.B * self.S):  # For each token
            token_input = flat_x[i]
            combined_expert_output = jnp.zeros(self.D, dtype=jnp.bfloat16)

            for k in range(self.K):  # For each chosen expert for that token
                expert_idx = flat_experts[i, k]
                weight = flat_weights[i, k]

                # Get kernels from the *full* gathered arrays ---
                gating_kernel = gating_kernel_full[expert_idx]
                up_proj_kernel = up_proj_kernel_full[expert_idx]
                down_proj_kernel = down_proj_kernel_full[expert_idx]

                # Perform the expert computation (dense matmuls)
                gating_proj = jnp.dot(token_input, gating_kernel)
                up_proj = jnp.dot(token_input, up_proj_kernel)

                # Note: Assuming 'silu' activation as specified in MoE init
                fused = nnx.silu(gating_proj) * up_proj

                expert_output = jnp.dot(fused, down_proj_kernel)

                # Apply router weight after computation (matches implementation)
                combined_expert_output += weight * expert_output

            expected_output = expected_output.at[i].set(combined_expert_output)

        expected_output = expected_output.reshape(self.B * self.S, self.D)

        # --- 3. Compare the results ---
        self.assertTrue(
            jnp.allclose(actual_output, expected_output, atol=1e-2, rtol=1e-2),
            f"The output of the distributed MoE does not match the dense equivalent.\n"
            f"Actual:\n{actual_output}\n"
            f"Expected:\n{expected_output}")
        print(
            "\n✅ Test Passed: Distributed MoE output matches the dense ground truth."
        )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock

import jax
import jax.numpy as jnp
import numpy as np
import pytest
from jax.sharding import Mesh

from tpu_inference.layers.common.attention_interface import attention
from tpu_inference.layers.common.attention_metadata import AttentionMetadata
from tpu_inference.runner.kv_cache import get_kv_cache_shape_with_mesh

# ---- Test Configuration & Constants ----

# Total number of tokens across all sequences in the batch
TOTAL_TOKENS = 10
# Number of sequences in the batch
NUM_SEQS = 2
# Padded maximum number of sequences
MAX_NUM_SEQS = 4
# Number of attention heads (Query)
NUM_HEADS = 8
# Number of attention heads (Key/Value) - for Grouped-Query Attention
NUM_KV_HEADS = 4
# Total number of blocks in the KV cache
NUM_BLOCKS = 32
# Number of tokens per block
BLOCK_SIZE = 16
# Maximum number of blocks a single sequence can occupy
MAX_BLOCKS_PER_SEQ = 8


@pytest.fixture
def mesh():
    """Provides a mock 1D JAX mesh for testing."""
    # Create a mesh with available devices, useful for running on CPU/GPU/TPU
    # For this test, it will likely be a single CPU device.
    devices = np.array(jax.local_devices()[:1])
    if not devices.any():
        # Add a mock device if no devices are present (e.g., in a CI environment)
        devices = np.array([jax.devices("cpu")[0]])
    return Mesh(devices.reshape((-1, 1, 1)), ("data", "attn_dp", "model"))


# ---- Test for `attention` ----


def _test_attention(monkeypatch, mesh, head_dim, use_sinks=False):
    """
    Tests the main `attention` function.

    Verifies that:
    1. It calls the `sharded_ragged_paged_attention` kernel with correct metadata.
    2. The final outputs (kv_cache and attention output) have the correct shapes.
    """
    # 1. Arrange

    # Create input tensors
    q_dtype = jnp.float32
    kv_dtype = jnp.float32
    q = jnp.ones((TOTAL_TOKENS, NUM_HEADS, head_dim), dtype=q_dtype)
    k = jnp.ones((TOTAL_TOKENS, NUM_KV_HEADS, head_dim), dtype=kv_dtype)
    v = jnp.ones((TOTAL_TOKENS, NUM_KV_HEADS, head_dim), dtype=kv_dtype)
    sinks = jnp.ones((NUM_HEADS, ), dtype=jnp.float32) if use_sinks else None

    kv_cache_shape = get_kv_cache_shape_with_mesh(
        mesh,
        NUM_BLOCKS,
        BLOCK_SIZE,
        NUM_KV_HEADS,
        head_dim,
        kv_dtype,
    )
    kv_cache = jnp.zeros(kv_cache_shape, dtype=kv_dtype)

    # Mock ragged_paged_attention to return a tensor of the correct shape
    mock_paged_attn_kernel = MagicMock(return_value=(jnp.ones(
        (TOTAL_TOKENS, NUM_HEADS, head_dim)), kv_cache), )

    if head_dim == 64:
        monkeypatch.setattr(
            "tpu_inference.layers.common.attention_interface.ragged_paged_attention_hd64",
            mock_paged_attn_kernel,
        )
    else:
        monkeypatch.setattr(
            "tpu_inference.layers.common.attention_interface.ragged_paged_attention",
            mock_paged_attn_kernel,
        )

    # Create AttentionMetadata
    attention_metadata = AttentionMetadata(
        input_positions=jnp.arange(TOTAL_TOKENS, dtype=jnp.int32),
        block_tables=jnp.zeros((MAX_NUM_SEQS * MAX_BLOCKS_PER_SEQ, ),
                               dtype=jnp.int32),
        seq_lens=jnp.array([5, 5, 0, 0], dtype=jnp.int32),
        query_start_loc=jnp.array([0, 5, 10, 10, 10], dtype=jnp.int32),
        request_distribution=jnp.array([0, 0, NUM_SEQS], dtype=jnp.int32),
    )

    # 2. Act
    final_kv_cache, output = attention(
        kv_cache=kv_cache,
        q=q,
        k=k,
        v=v,
        attention_metadata=attention_metadata,
        mesh=mesh,
        head_dim_original=head_dim,
        sinks=sinks,
    )

    # 3. Assert
    # Check that both mocked kernels were called
    mock_paged_attn_kernel.assert_called_once()

    # Check output shapes
    assert final_kv_cache.shape == kv_cache.shape
    assert output.shape == q.shape

    # Check that the output is the one from our mock
    assert jnp.all(output == 1.0)


def test_attention(monkeypatch, mesh):
    _test_attention(monkeypatch, mesh, 128)


def test_attention_hd64(monkeypatch, mesh):
    _test_attention(monkeypatch, mesh, 64)


def test_attention_sink(monkeypatch, mesh):
    _test_attention(monkeypatch, mesh, 64, True)


def test_attention_sink_no_64_raises_error(monkeypatch, mesh):
    with pytest.raises(
            NotImplementedError,
            match="Attention sink support is only available when head_dim==64"
    ):
        _test_attention(monkeypatch, mesh, 128, True)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

import jax
import torch
from jax.sharding import PartitionSpec
from vllm.model_executor.layers.fused_moe.layer import FusedMoE
from vllm.model_executor.layers.linear import LinearBase, LinearMethodBase
from vllm.model_executor.layers.quantization import \
    register_quantization_config
from vllm.model_executor.layers.quantization.base_config import \
    QuantizeMethodBase
from vllm.model_executor.layers.quantization.fp8 import (Fp8Config,
                                                         Fp8LinearMethod)
from vllm.model_executor.layers.quantization.utils.quant_utils import \
    is_layer_skipped

from tpu_inference.layers.common.quant_methods import FP8, get_tpu_quant_method
from tpu_inference.layers.vllm.quantization.configs import (
    VllmQuantConfig, VllmQuantLinearConfig)
from tpu_inference.layers.vllm.quantization.unquantized import \
    VllmUnquantizedLinearMethod
from tpu_inference.logger import init_logger

P = PartitionSpec

logger = init_logger(__name__)


@register_quantization_config(get_tpu_quant_method(FP8))
class VllmFp8Config(Fp8Config, VllmQuantConfig):

    @classmethod
    def get_name(cls):
        return FP8

    def get_supported_act_dtypes(self) -> list[torch.dtype]:
        return [torch.bfloat16]

    def get_quant_method(
        self, layer: torch.nn.Module, prefix: str
    ) -> Optional[Union["LinearMethodBase", "QuantizeMethodBase"]]:
        if isinstance(layer, LinearBase):
            linear_config = self.get_linear_config(layer)
            if is_layer_skipped(prefix, self.ignored_layers):
                return VllmUnquantizedLinearMethod(linear_config)
            return VllmFp8LinearMethod(self, linear_config)
        elif isinstance(layer, FusedMoE):
            raise NotImplementedError(
                "FP8 FusedMoE is currently not supported in torchax-jax")
        return None


class VllmFp8LinearMethod(Fp8LinearMethod):

    def __init__(self, quant_config: VllmFp8Config,
                 jax_config: VllmQuantLinearConfig):
        super().__init__(quant_config)
        self.jax_config = jax_config
        self._configure_sharding()

    def _configure_sharding(self) -> None:

        raise NotImplementedError(
            "Configure PartitionSpec for weight_sharding and scale_sharding "
            "based on layer type (RowParallel/ColumnParallel)")

    def process_weights_after_loading(self, layer: torch.nn.Module) -> None:

        raise NotImplementedError(
            "Convert layer.weight, layer.weight_scale, and optionally "
            "layer.input_scale and layer.bias from torch tensors to JAX arrays "
            "using torch_to_jax_param() with appropriate sharding")

    def apply(self,
              layer: torch.nn.Module,
              x: torch.Tensor,
              bias: Optional[torch.Tensor] = None) -> torch.Tensor:

        with jax.named_scope(layer._get_name()):
            if self.jax_config.fuse_matmuls:
                out = self._apply_fused(layer, x, bias)
            else:
                out = self._apply_split(layer, x, bias)

        return out

    def _apply_fused(self,
                     layer: torch.nn.Module,
                     x: torch.Tensor,
                     bias: Optional[torch.Tensor] = None) -> torch.Tensor:

        raise NotImplementedError(
            "Implement single matmul for fused outputs: "
            "quantize input to fp8, perform fp8 matmul with weight and scales, "
            "dequantize output, and add bias if present")

    def _apply_split(self,
                     layer: torch.nn.Module,
                     x: torch.Tensor,
                     bias: Optional[torch.Tensor] = None) -> torch.Tensor:

        raise NotImplementedError(
            "Implement separate matmuls per output partition: "
            "split weight/scale by output_sizes, perform fp8 matmul for each, "
            "concatenate results, and add bias if present")



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum

import jax
import jax.numpy as jnp
import torch
from jax.sharding import Mesh
from torchax.interop import jax_view, torch_view
from vllm.model_executor.layers.fused_moe.config import FusedMoEConfig
from vllm.model_executor.layers.fused_moe.layer import FusedMoE

from tpu_inference import envs
from tpu_inference.kernels.fused_moe.v1.kernel import fused_ep_moe
from tpu_inference.layers.common.fused_moe_gmm import fused_moe_func
from tpu_inference.logger import init_logger

logger = init_logger(__name__)


class FusedMoEBackend(Enum):
    FUSED_MOE = "fused_moe"
    GMM_EP = "gmm_ep"
    GMM_TP = "gmm_tp"


def select_moe_backend(moe: FusedMoEConfig):
    if envs.USE_MOE_EP_KERNEL:
        if moe.use_ep:
            return FusedMoEBackend.FUSED_MOE
        logger.warning_once(
            "USE_MOE_EP_KERNEL=1 but expert parallelism is not "
            "enabled. Falling back to gmm implementation.")

    if moe.use_ep:
        return FusedMoEBackend.GMM_EP

    # Use default implementation.
    return FusedMoEBackend.GMM_TP


def fused_moe_apply(
    layer: torch.nn.Module,
    x: torch.Tensor,
    router_logits: torch.Tensor,
    moe_backend: FusedMoEBackend,
    mesh: Mesh,
    extra_backend_kwargs: dict,
) -> torch.Tensor:
    assert isinstance(layer, FusedMoE)
    if layer.scoring_func != "softmax":
        raise NotImplementedError("Only softmax is supported for scoring_func")

    x = jax_view(x)
    gating_output = jax_view(router_logits)

    w13_weight = jax_view(layer.w13_weight)
    w13_weight_scale = jax_view(getattr(layer, "w13_weight_scale", None))
    w13_bias = jax_view(getattr(layer, "w13_bias", None))
    w2_weight = jax_view(layer.w2_weight)
    w2_weight_scale = jax_view(getattr(layer, "w2_weight_scale", None))
    w2_bias = jax_view(getattr(layer, "w2_bias", None))

    with jax.named_scope(layer._get_name()):
        match moe_backend:
            case FusedMoEBackend.FUSED_MOE:
                actual_hidden_size = x.shape[-1]
                padding_size = w13_weight.shape[-2] - actual_hidden_size
                x = jnp.pad(x, ((0, 0), (0, padding_size)))
                output = fused_ep_moe(
                    mesh=mesh,
                    tokens=x,
                    w1=w13_weight,
                    w2=w2_weight,
                    w1_scale=w13_weight_scale,
                    w2_scale=w2_weight_scale,
                    b1=w13_bias,
                    b2=w2_bias,
                    gating_output=gating_output,
                    top_k=layer.top_k,
                    renormalize_topk_logits=layer.renormalize,
                    act_fn=layer.activation,
                    **extra_backend_kwargs,
                )[:, :actual_hidden_size]
            case FusedMoEBackend.GMM_EP | FusedMoEBackend.GMM_TP:
                output = fused_moe_func(
                    hidden_states=x,
                    w1=w13_weight,
                    w2=w2_weight,
                    w1_scale=w13_weight_scale,
                    w2_scale=w2_weight_scale,
                    w1_bias=w13_bias,
                    w2_bias=w2_bias,
                    gating_output=gating_output,
                    topk=layer.top_k,
                    renormalize=layer.renormalize,
                    mesh=mesh,
                    use_ep=layer.use_ep,
                    activation=layer.activation,
                )

        return torch_view(output)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile
from typing import Optional

import jax
import pytest
import torch
import torchax
from jax.sharding import NamedSharding, PartitionSpec
from torchax.interop import torch_view
from torchax.ops.mappings import j2t, t2j
from vllm.config import set_current_vllm_config
from vllm.distributed.parallel_state import (ensure_model_parallel_initialized,
                                             init_distributed_environment)
from vllm.engine.arg_utils import EngineArgs
from vllm.model_executor.layers.linear import (ColumnParallelLinear,
                                               LinearBase,
                                               MergedColumnParallelLinear,
                                               QKVParallelLinear,
                                               RowParallelLinear)
from vllm.model_executor.layers.quantization.compressed_tensors.compressed_tensors import \
    CompressedTensorsLinearMethod
from vllm.model_executor.model_loader import get_model as vllm_get_model

from tpu_inference.layers.vllm.quantization import get_tpu_quantization_config
from tpu_inference.layers.vllm.quantization.compressed_tensors.compressed_tensors import \
    VllmCompressedTensorsConfig
from tpu_inference.layers.vllm.quantization.compressed_tensors.schemes.compressed_tensors_w8a8_int8 import \
    VllmCompressedTensorsW8A8Int8

from . import utils as test_utils

P = PartitionSpec
MODELS = ["RedHatAI/Qwen2.5-1.5B-quantized.w8a8"]


def ref_quantize_int8(x: torch.Tensor):
    x_abs_max = torch.amax(torch.abs(x), dim=1, keepdim=True)
    x_s = x_abs_max / 127
    x_q = torch.round(x / x_s).to(torch.int8)
    return x_q, x_s.to(torch.float32)


def ref_w8a8_int8(x: torch.Tensor, w_q: torch.Tensor, w_s: torch.Tensor,
                  b: Optional[torch.Tensor]):
    x_q, x_s = ref_quantize_int8(x)
    out = torch.einsum('bd,fd->bf', x_q.to(torch.float32),
                       w_q.to(torch.float32))
    out = (out * x_s) * w_s.T
    if b is not None:
        out += b
    return out.to(x.dtype)


@pytest.fixture(autouse=True)
def setup_environment():
    # This is a fake config used for init dist env.
    # RowParallelLinear needs dist env to be initialized.
    engine_args = EngineArgs(
        model=MODELS[0],
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )

    vllm_config = engine_args.create_engine_config()

    with set_current_vllm_config(vllm_config):
        temp_file = tempfile.mkstemp()[1]
        init_distributed_environment(
            1,
            0,
            local_rank=0,
            distributed_init_method=f"file://{temp_file}",
            backend="gloo")
        ensure_model_parallel_initialized(1, 1)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
def test_quant_override(model, mesh):

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.model_config.dtype = torch.bfloat16

    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    assert isinstance(quant_config, VllmCompressedTensorsConfig)
    assert quant_config.vllm_config == vllm_config
    assert quant_config.mesh == mesh


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("mesh", [
    test_utils.get_spmd_mesh(1),
    test_utils.get_spmd_mesh(jax.local_device_count())
])
def test_loading_model(model, mesh):
    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.model_config.dtype = torch.bfloat16
    vllm_config.quant_config = get_tpu_quantization_config(vllm_config, mesh)
    vllm_config.device_config.device = "cpu"

    vllm_model = vllm_get_model(vllm_config=vllm_config)
    layers = test_utils.find_all_layer_type(vllm_model, LinearBase)
    for layer in layers:
        assert isinstance(layer.quant_config, VllmCompressedTensorsConfig)
        assert isinstance(layer.quant_method, CompressedTensorsLinearMethod)
        assert isinstance(layer.scheme, VllmCompressedTensorsW8A8Int8)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_row_parallel_linear(model, bias, num_devices, enable_sp,
                             enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)

    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    # Call tpu_inference code
    vllm_config.model_config.dtype = dtype
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        jax_row_linear = RowParallelLinear(
            input_size=4096,
            output_size=8192,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    weight_data_float = torch.rand(
        (jax_row_linear.output_size, jax_row_linear.input_size),
        dtype=dtype) / 10
    weight_data, weight_scale_data = ref_quantize_int8(weight_data_float)
    if bias:
        bias_data = torch.rand_like(jax_row_linear.bias.data)

    jax_row_linear.weight.data = weight_data
    jax_row_linear.weight_scale.data = weight_scale_data
    if bias:
        jax_row_linear.bias.data = bias_data

    input_tensor = torch.rand(10, jax_row_linear.input_size, dtype=dtype) / 10
    input_tensor = input_tensor.to('cpu')

    jax_input_tensor = torch_view(t2j(input_tensor, use_dlpack=False))
    jax_input_tensor.apply_jax_(jax.device_put,
                                NamedSharding(mesh, P(None, None)))
    with torchax.default_env():
        assert isinstance(jax_row_linear.quant_method,
                          CompressedTensorsLinearMethod)
        assert isinstance(jax_row_linear.scheme, VllmCompressedTensorsW8A8Int8)
        jax_row_linear.quant_method.process_weights_after_loading(
            jax_row_linear)
        jax_output = jax_row_linear(jax_input_tensor)
        jax_output = j2t(jax_output.to(torch.float32)).to(dtype)

    # Call reference w8a8 int8
    output = ref_w8a8_int8(
        input_tensor,
        weight_data,
        weight_scale_data,
        bias_data if bias else None,
    )

    torch.testing.assert_close(output, jax_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_column_parallel_linear(model, bias, num_devices, enable_sp,
                                enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)
    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    # Call tpu_inference code
    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        jax_column_linear = ColumnParallelLinear(
            input_size=4096,
            output_size=8192,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )

    weight_data_float = torch.rand(
        (jax_column_linear.output_size, jax_column_linear.input_size),
        dtype=dtype) / 10
    weight_data, weight_scale_data = ref_quantize_int8(weight_data_float)
    if bias:
        bias_data = torch.rand_like(jax_column_linear.bias.data)

    jax_column_linear.weight.data = weight_data
    jax_column_linear.weight_scale.data = weight_scale_data
    if bias:
        jax_column_linear.bias.data = bias_data

    input_tensor = torch.rand(10, jax_column_linear.input_size,
                              dtype=dtype) / 10
    input_tensor = input_tensor.to('cpu')

    jax_input_tensor = torch_view(t2j(input_tensor, use_dlpack=False))
    jax_input_tensor.apply_jax_(jax.device_put,
                                NamedSharding(mesh, P(None, None)))
    with torchax.default_env():
        assert isinstance(jax_column_linear.quant_method,
                          CompressedTensorsLinearMethod)
        assert isinstance(jax_column_linear.scheme,
                          VllmCompressedTensorsW8A8Int8)
        jax_column_linear.quant_method.process_weights_after_loading(
            jax_column_linear)
        jax_output = jax_column_linear(jax_input_tensor)
        jax_output = j2t(jax_output.to(torch.float32)).to(dtype)

    # Call reference w8a8 int8
    output = ref_w8a8_int8(
        input_tensor,
        weight_data,
        weight_scale_data,
        bias_data if bias else None,
    )

    torch.testing.assert_close(output, jax_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("fuse_matmuls", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_qkv_parallel_linear(model, bias, num_devices, enable_sp, fuse_matmuls,
                             enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)
    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    # Call tpu_inference code
    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        jax_qkv_linear = QKVParallelLinear(
            hidden_size=4096,
            head_size=128,
            total_num_heads=32,
            total_num_kv_heads=8,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )
        jax_qkv_linear.quant_method.fuse_matmuls = fuse_matmuls

    weight_data_float = torch.rand(
        (jax_qkv_linear.output_size, jax_qkv_linear.input_size),
        dtype=dtype) / 10
    weight_data, weight_scale_data = ref_quantize_int8(weight_data_float)
    if bias:
        bias_data = torch.rand_like(jax_qkv_linear.bias.data)

    jax_qkv_linear.weight.data = weight_data
    jax_qkv_linear.weight_scale.data = weight_scale_data
    if bias:
        jax_qkv_linear.bias.data = bias_data

    input_tensor = torch.rand(10, jax_qkv_linear.input_size, dtype=dtype) / 10
    input_tensor = input_tensor.to('cpu')

    jax_input_tensor = torch_view(t2j(input_tensor, use_dlpack=False))
    jax_input_tensor.apply_jax_(jax.device_put,
                                NamedSharding(mesh, P(None, None)))
    with torchax.default_env():
        assert isinstance(jax_qkv_linear.quant_method,
                          CompressedTensorsLinearMethod)
        assert isinstance(jax_qkv_linear.scheme, VllmCompressedTensorsW8A8Int8)
        jax_qkv_linear.quant_method.process_weights_after_loading(
            jax_qkv_linear)
        jax_output = jax_qkv_linear(jax_input_tensor)
        jax_output = j2t(jax_output.to(torch.float32)).to(dtype)

    # Call reference w8a8 int8
    output = ref_w8a8_int8(
        input_tensor,
        weight_data,
        weight_scale_data,
        bias_data if bias else None,
    )

    torch.testing.assert_close(output, jax_output)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("bias", [False, True])
@pytest.mark.parametrize("num_devices", [1, jax.local_device_count()])
@pytest.mark.parametrize("fuse_matmuls", [False, True])
@pytest.mark.parametrize("enable_sp", [False, True])
@pytest.mark.parametrize("enable_attn_dp", [False, True])
def test_merged_column_parallel_linear(model, bias, num_devices, fuse_matmuls,
                                       enable_sp, enable_attn_dp):
    # Skip if enable_attn_dp is True but we don't have enough devices
    if enable_attn_dp and num_devices < 2:
        pytest.skip("enable_attn_dp requires at least 2 devices")

    mesh = test_utils.get_spmd_mesh(num_devices, enable_attn_dp)
    dtype = torch.bfloat16

    engine_args = EngineArgs(
        model=model,
        max_model_len=64,
        max_num_batched_tokens=64,
        max_num_seqs=4,
    )
    vllm_config = engine_args.create_engine_config()
    vllm_config.compilation_config.pass_config.enable_sp = enable_sp

    # Call tpu_inference code
    vllm_config.model_config.dtype = torch.bfloat16
    quant_config = get_tpu_quantization_config(vllm_config, mesh)
    with set_current_vllm_config(vllm_config):
        jax_merged_column_linear = MergedColumnParallelLinear(
            input_size=4096,
            output_sizes=[14336] * 2,
            bias=bias,
            params_dtype=dtype,
            return_bias=False,
            quant_config=quant_config,
        )
        jax_merged_column_linear.quant_method.fuse_matmuls = fuse_matmuls

    weight_data_float = torch.rand((jax_merged_column_linear.output_size,
                                    jax_merged_column_linear.input_size),
                                   dtype=dtype) / 10
    weight_data, weight_scale_data = ref_quantize_int8(weight_data_float)
    if bias:
        bias_data = torch.rand_like(jax_merged_column_linear.bias.data)

    jax_merged_column_linear.weight.data = weight_data
    jax_merged_column_linear.weight_scale.data = weight_scale_data
    if bias:
        jax_merged_column_linear.bias.data = bias_data

    input_tensor = torch.rand(
        10, jax_merged_column_linear.input_size, dtype=dtype) / 10
    input_tensor = input_tensor.to('cpu')

    jax_input_tensor = torch_view(t2j(input_tensor, use_dlpack=False))
    jax_input_tensor.apply_jax_(jax.device_put,
                                NamedSharding(mesh, P(None, None)))
    with torchax.default_env():
        assert isinstance(jax_merged_column_linear.quant_method,
                          CompressedTensorsLinearMethod)
        assert isinstance(jax_merged_column_linear.scheme,
                          VllmCompressedTensorsW8A8Int8)
        jax_merged_column_linear.quant_method.process_weights_after_loading(
            jax_merged_column_linear)
        jax_output = jax_merged_column_linear(jax_input_tensor)
        jax_output = j2t(jax_output.to(torch.float32)).to(dtype)

    # Call reference w8a8 int8
    output = ref_w8a8_int8(
        input_tensor,
        weight_data,
        weight_scale_data,
        bias_data if bias else None,
    )

    torch.testing.assert_close(output, jax_output)
